from typing import NamedTuple, Optional

import dash
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from authlib.oidc.core import UserInfo
from dash import State, html
from dash.development.base_component import Component
from dash_extensions.enrich import Input, Output, callback
from flask import session

from amora.dash.components import user_avatar


class NavItem(NamedTuple):
    fa_icon: str
    href: str
    title: str


def nav() -> dbc.Nav:
    return dbc.Nav(
        [
            dbc.NavLink(
                [
                    html.I(className=f"fa-solid {page.get('fa_icon')}"),
                    " ",
                    page["name"],
                ],
                href=page["relative_path"],
                active="exact",
            )
            for page in dash.page_registry.values()
            if page.get("location") == "sidebar"
        ],
        vertical=True,
        pills=True,
    )


layout_id = "side-bar"


def layout() -> Component:
    return dbc.Offcanvas(
        [
            dmc.Center(html.H2("🌱", className="display-4")),
            dmc.Center(html.H2("Amora", className="display-4")),
            dmc.Space(h="lg"),
            nav(),
            dmc.Space(h=60),
            html.Hr(),
            dmc.Center(id="user-account"),
            dmc.Center(
                dbc.Alert(
                    ["Press ", dmc.Kbd("m"), " to toggle the menu"], color="light"
                )
            ),
        ],
        id=layout_id,
        is_open=True,
    )


@callback(Output("user-account", "children"), [Input("url", "pathname")])
def display_user(pathname):
    user: Optional[UserInfo] = session.get("user")
    if not user:
        return dbc.Button(
            children=[html.I(className=f"fa-solid fa-user"), " Login"],
            href="/login",
            external_link=True,
        )
    else:
        return user_avatar.layout(user["userinfo"])


@callback(
    Output(layout_id, "is_open"),
    Input("event-listener", "n_events"),
    [State(layout_id, "is_open"), State("event-listener", "event")],
)
def toggle_menu_on_keydown(
    _n_events: Optional[int], is_open: bool, event: Optional[dict]
) -> bool:
    if event and event["key"] == "m":
        return not is_open
    return is_open



from dash.development.base_component import Component

from amora.dash.components.filters import accepted_values_filter, date_filter
from amora.dashboards import AcceptedValuesFilter, DateFilter, Filter


def layout(filter: Filter) -> Component:
    if isinstance(filter, DateFilter):
        return date_filter.layout(filter)

    if isinstance(filter, AcceptedValuesFilter):
        return accepted_values_filter.layout(filter)



import os
from pathlib import Path
from typing import Iterable, Union

from amora.config import settings


def list_files(path: Union[str, Path], suffix: str) -> Iterable[Path]:
    yield from Path(path).rglob(f"*{suffix}")


def model_path_for_target_path(path: Path) -> Path:
    return Path(
        str(path)
        .replace(settings.target_path.as_posix(), settings.models_path.as_posix())
        .replace(".sql", ".py"),
    )


def target_path_for_model_path(path: Path) -> Path:
    return Path(
        str(path)
        .replace(settings.models_path.as_posix(), settings.target_path.as_posix())
        .replace(".py", ".sql")
    )


def clean_compiled_files() -> None:
    for sql_file in list_target_files():
        os.remove(sql_file)


def list_target_files() -> Iterable[Path]:
    return list_files(settings.target_path, suffix=".sql")



from sqlalchemy import Date, DateTime, String, Time, func
from sqlalchemy.sql.functions import Function

from amora.models import ColumnElement
from amora.providers.bigquery import TimePart


def remove_non_numbers(column: ColumnElement) -> Function:
    """
    The column string value with numeric characters only.

     E.g: "31.752.270/0001-82" -> "31752270000182"
    """
    return func.regexp_replace(column, "[^0-9]", "", type_=String)


def remove_leading_zeros(column: ColumnElement) -> Function:
    """
    The column string value without leading zeros.

     E.g: "00001000000" -> "1000000"
    """
    return func.regexp_replace(column, "^0+", "", type_=String)


def parse_numbers(column: ColumnElement) -> Function:
    """
    Parses a string column as a number, returning NULL if value contains 0 numbers

    E.g: "0031.752.270/0001-82" -> "31752270000182"
         "IM_A_STRING_WITH_NO_NUMBERS" -> NULL
    """
    return func.nullif(
        remove_leading_zeros(remove_non_numbers(column)), "", type_=String
    )


def datetime_trunc_hour(column: ColumnElement) -> Function:
    """
    Truncate a datetime column by its HOUR part.
    "2019-12-09T16:15:20" -> "2019-12-09T16:00:00"

    E.g.: `datetime_trunc_hour(Steps.creationDate)`

    ```sql
    DATETIME(
        DATE(`steps`.`creationDate`),
        TIME_TRUNC(TIME(`steps`.`creationDate`), HOUR)
    )
    ```
    """
    date_part = func.date(column, type_=Date)
    hour_part = func.time_trunc(func.time(column), TimePart.HOUR.value, type_=Time)
    return func.datetime(date_part, hour_part, type_=DateTime)



import dash_bootstrap_components as dbc
from dash import html
from dash.development.base_component import Component

from amora.models import Model


def component(model: Model) -> Component:
    return html.Span(
        [dbc.Badge(str(label), color="info") for label in model.__model_config__.labels]
    )



import os
from datetime import datetime
from typing import Iterable, Optional

from sqlalchemy import func

from amora.config import settings as default_settings
from amora.models import (
    AmoraModel,
    Field,
    MaterializationTypes,
    ModelConfig,
    Session,
    select,
)
from amora.protocols import Compilable
from amora.storage import local_engine, local_metadata
from amora.version import VERSION


class AuditLog(AmoraModel, table=True):
    __model_config__ = ModelConfig(
        materialized=MaterializationTypes.table,
        description="Stores test log data",
    )
    metadata = local_metadata

    test_run_id: str = Field(
        primary_key=True,
        description="Unique id of the test run",
        nullable=False,
    )
    test_node_id: str = Field(
        primary_key=True,
        description="pytest full node id of the item",
        nullable=False,
    )
    bytes_billed: int = Field(
        description="Total billable scanned bytes during query executions"
    )
    estimated_cost_in_usd: float = Field(
        description="Estimated cost of query executions"
    )
    query: str = Field(description="SQL query executed for data assertion")
    user_email: Optional[str] = Field(
        description="GCP user email that performed the query", nullable=True
    )
    execution_time_in_ms: int = Field(description="Query execution time in miliseconds")
    inserted_at: datetime = Field(
        default_factory=datetime.utcnow,
        description="UTC Datetime of the insert",
    )
    referenced_tables: str = Field(
        description="JSON encoded referenced tables on the query", nullable=True
    )
    settings: str = Field(
        description="JSON encoded current`amora.config.settings`",
        default=default_settings.json(),
    )
    amora_version: str = Field(
        description="Current version of the amora package", default=VERSION
    )
    xdist_worker_id: Optional[str] = Field(
        description="The name of the `pytest-xdist` worker. E.g.: gw2",
        default=os.getenv("PYTEST_XDIST_WORKER"),
    )

    @classmethod
    def get_all(cls, test_run_id: str) -> Iterable["AuditLog"]:
        with Session(local_engine) as session:
            statement = select(cls).where(cls.test_run_id == test_run_id)
            return (log for (log, *_) in session.exec(statement).all())  # type: ignore


class AuditReport(AmoraModel, table=True):
    __depends_on__ = [AuditLog]
    __model_config__ = ModelConfig(
        materialized=MaterializationTypes.view,
        description="Stores test run reports",
    )
    metadata = local_metadata

    test_run_id: str = Field(primary_key=True)
    amora_version: str = Field(
        description="Current version of the amora package", default=VERSION
    )
    user_email: Optional[str] = Field(
        description="GCP user email that performed the query", nullable=True
    )

    total_query_time: int
    total_cost: float
    total_bytes_billed: int

    @classmethod
    def source(cls) -> Optional[Compilable]:
        raise NotImplementedError  # WIP

        return select(
            AuditLog.test_run_id,
            AuditLog.amora_version,
            AuditLog.user_email,
            func.sum(AuditLog.execution_time_in_ms).label(cls.total_query_time.key),
            func.sum(AuditLog.estimated_cost_in_usd).label(cls.total_cost.key),
            func.sum(AuditLog.bytes_billed).label(cls.total_bytes_billed.key),
        ).group_by(AuditLog.test_run_id, AuditLog.amora_version, AuditLog.user_email)


AuditLog.__table__.create(bind=local_engine, checkfirst=True)



import dash
import dash_bootstrap_components as dbc
from dash_extensions import EventListener
from dash_extensions.enrich import Dash, dcc

from amora.dash.authentication import add_auth0_login
from amora.dash.components import side_bar
from amora.dash.config import settings
from amora.models import list_models

dash_app = Dash(
    __name__, external_stylesheets=settings.external_stylesheets, use_pages=True
)

if settings.auth0_login_enabled:
    add_auth0_login(dash_app)

LISTENABLE_EVENTS = [
    {
        "event": "keydown",
        "props": ["key", "altKey", "ctrlKey", "shiftKey", "metaKey"],
    },
]
# App
dash_app.layout = dbc.Container(
    children=[
        dcc.Location(id="url"),
        EventListener(events=LISTENABLE_EVENTS, logging=True, id="event-listener"),
        side_bar.layout(),
        dbc.Row(
            dash.page_container,
            id="page-content",
        ),
    ],
)

# fixme: should be in a post init callback or shouldn't be needed at all
list(list_models())



class Lotties:
    site_analytics_dashboard = (
        "https://assets1.lottiefiles.com/packages/lf20_i2eyukor.json"
    )
    analytics_people = "https://assets7.lottiefiles.com/packages/lf20_dews3j6m.json"
    green_bi_dashboard = "https://assets8.lottiefiles.com/packages/lf20_mcvtkrvc.json"
    notebook_browser_analytics = (
        "https://assets6.lottiefiles.com/packages/lf20_4ejZJA.json"
    )
    metaverse_explorer = (
        "https://assets3.lottiefiles.com/private_files/lf30_fhbalxer.json"
    )
    loading = "https://assets7.lottiefiles.com/packages/lf20_poqmycwy.json"
    loading_pacman = "https://assets2.lottiefiles.com/packages/lf20_lit5uqwc.json"



from enum import Enum
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, Optional

from feast.usage import USAGE_ENDPOINT as FEAST_USAGE_ENDPOINT
from pydantic import BaseSettings

from amora.config import ROOT_PATH


class FeatureStoreProviders(str, Enum):
    local = "local"
    gcp = "gcp"


class FeatureStoreOnlineStoreTypes(str, Enum):
    redis = "redis"
    sqlite = "sqlite"
    datastore = "datastore"


class FeatureStoreOfflineStoreTypes(str, Enum):
    bigquery = "bigquery"
    file = "file"


class FeatureStoreSettings(BaseSettings):
    REGISTRY: str = NamedTemporaryFile(
        suffix="amora-feature-store-registry", delete=False
    ).name
    REPO_PATH: str = NamedTemporaryFile(suffix="repo-path", delete=False).name
    PROVIDER: str = FeatureStoreProviders.local.value
    OFFLINE_STORE_TYPE: str = FeatureStoreOfflineStoreTypes.file.value
    OFFLINE_STORE_CONFIG: Dict[str, str] = {}

    ONLINE_STORE_TYPE: str = FeatureStoreOnlineStoreTypes.sqlite.value
    ONLINE_STORE_CONFIG: Dict[str, str] = {
        "path": Path(ROOT_PATH).joinpath("amora-online-feature-store.db").name
    }
    DEFAULT_FEATURE_TTL_IN_SECONDS: int = 3600

    HTTP_SERVER_HOST: str = "0.0.0.0"
    HTTP_SERVER_PORT: int = 8666
    HTTP_ACCESS_LOG_ENABLED: bool = False

    TQDM_ASCII_LOGGING: bool = False
    TQDM_DISABLE: Optional[bool] = None

    USAGE_TRACKING_ENABLED: bool = False
    USAGE_ENDPOINT: str = FEAST_USAGE_ENDPOINT

    class Config:
        env_prefix = "AMORA_FEATURE_STORE_"


settings = FeatureStoreSettings()



import dash_bootstrap_components as dbc
from dash import html
from dash.development.base_component import Component

from amora.dag import DependencyDAG
from amora.dash.components import (
    dependency_dag,
    materialization_type_badge,
    model_code,
    model_columns,
    model_datatable,
    model_summary,
)
from amora.models import Model


def component(model: Model) -> Component:
    model_config = model.__model_config__
    return dbc.Card(
        [
            dbc.CardHeader(
                html.H4(model.unique_name(), className="card-title"),
            ),
            dbc.CardBody(
                [
                    dependency_dag.component(DependencyDAG.from_model(model)),
                    materialization_type_badge.component(model_config.materialized),
                    html.P(
                        model_config.description,
                        className="card-text",
                    ),
                    dbc.Accordion(
                        [
                            dbc.AccordionItem(
                                model_summary.component(model),
                                title="📈 Summary",
                                item_id="summary",
                            ),
                            dbc.AccordionItem(
                                model_columns.component(model),
                                title="📝 Docs",
                                item_id="docs",
                            ),
                            dbc.AccordionItem(
                                model_datatable.component(model),
                                title="🍰 Sample dataset",
                                item_id="sample",
                            ),
                            dbc.AccordionItem(
                                model_code.python_component(model),
                                title="🐍 Python Code",
                                item_id="python_code",
                            ),
                            dbc.AccordionItem(
                                model_code.sql_component(model),
                                title="🏗 SQL Code",
                                item_id="sql_code",
                            ),
                        ],
                        start_collapsed=True,
                        id="model-details-accordion",
                    ),
                ]
            ),
        ],
    )



import dash_bootstrap_components as dbc

from amora.models import MaterializationTypes


def component(materialization_type: MaterializationTypes) -> dbc.Badge:
    return dbc.Badge(materialization_type.value, color="primary", className="me-1")



from feast import BigQuerySource, Feature, FeatureView
from google.protobuf.duration_pb2 import Duration

from amora.feature_store import settings
from amora.feature_store.protocols import FeatureViewSourceProtocol
from amora.feature_store.type_mapping import type_for_column
from amora.models import Model
from amora.providers.bigquery import get_fully_qualified_id


def name_for_model(model: Model) -> str:
    """
    Feature View Name is the name of the group of features.
    """
    # fixme: type ignore não deveria ser necessário aqui
    return model.__tablename__  # type: ignore


def feature_view_for_model(model: Model) -> FeatureView:
    """
    A feature view is an object that represents a logical group of time-series
    feature data as it is found in a model.
    """
    if not isinstance(model, FeatureViewSourceProtocol):
        raise ValueError(
            f"Feature view models (`@feature_view`) must implement the "
            f"{FeatureViewSourceProtocol.__name__} protocol. "
            f"{model} failed the check"
        )

    return FeatureView(
        name=name_for_model(model),
        entities=[col.name for col in model.feature_view_entities()],
        features=[
            Feature(
                name=col.name,
                dtype=type_for_column(col),
            )
            for col in model.feature_view_features()
        ],
        batch_source=BigQuerySource(
            table_ref=get_fully_qualified_id(model),
            event_timestamp_column=model.feature_view_event_timestamp().key,
        ),
        ttl=Duration(seconds=settings.DEFAULT_FEATURE_TTL_IN_SECONDS),
    )



"""
Tests that can be reused by multiple projects
"""
from amora.models import MaterializationTypes, list_models
from amora.providers.bigquery import (
    get_schema,
    schema_for_model,
    schema_for_model_source,
)


def test_materialized_schema_equal_local_schema():
    """
    Validates that the current materialization has the same schema
    as the locally defined model
    """
    for model, model_path in list_models():
        if MaterializationTypes.ephemeral == model.__model_config__.materialized:
            continue

        materialized_schema = get_schema(model.unique_name())
        model_schema = schema_for_model(model)

        assert set(materialized_schema) == set(model_schema), (
            f"Diff found between the materialized model `{model.unique_name()}` "
            f"schema and the local definition `{model_path}`"
        )


def test_models_schema_equal_its_source_schema():
    """
    Validates that the model schema matches its `source` classmethod result schema
    """
    for model, _model_path in list_models():
        source = model.source()

        if source is None:
            continue

        model_schema = schema_for_model(model)
        source_schema = schema_for_model_source(model)

        assert set(source_schema) == set(model_schema), (
            f"Diff found between the schema of `source` classmethod "
            f"and the model schema definition on `{model.unique_name()}`"
        )



import dash
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from dash import html
from dash.development.base_component import Component
from dash_extensions import Lottie

from amora.config import settings
from amora.dash.components.animation import Lotties

dash.register_page(__name__, path="/", fa_icon="fa-house", location="sidebar")


def layout() -> Component:
    return dbc.Row(
        [
            dmc.Center(html.H1(f"Project: {settings.TARGET_PROJECT}")),
            Lottie(
                options=dict(
                    loop=False,
                    autoplay=True,
                    rendererSettings=dict(preserveAspectRatio="xMidYMid slice"),
                ),
                width="75%",
                url=Lotties.site_analytics_dashboard,
            ),
        ]
    )



import dash_mantine_components as dmc
from authlib.oidc.core import UserInfo
from dash import dcc
from dash.development.base_component import Component

from amora.dash.config import Size


def layout(user_info: UserInfo, size: Size = Size.large) -> Component:
    return dcc.Link(
        dmc.Avatar(
            src=user_info["picture"],
            alt=user_info["given_name"],
            radius=Size.extra_large,
            size=size,
        ),
        href=f"/users/{user_info['email']}",
        title=user_info["name"],
    )



import pandas as pd
from dash import dash_table, html
from dash.development.base_component import Component

from amora.models import Model


def component(model: Model) -> Component:
    df = pd.DataFrame(
        [
            {
                "column": col.key,
                "description": col.comment if col.comment else "Undocumented",
            }
            for col in model.__table__.columns
        ]
    )
    return html.Div(
        dash_table.DataTable(
            data=df.to_dict("records"),
            columns=[
                {"name": "Column", "id": "column"},
                {"name": "Description", "id": "description"},
            ],
            style_cell={"textAlign": "left"},
            style_as_list_view=True,
        ),
        id=f"model-columns-{model.unique_name()}",
    )



from amora.feature_store.config import settings


def patch_tqdm() -> None:
    """
    Monkey patch tqdm logging to use only ASCII characters, instead of the default UNICODE format
    """
    from feast import feature_store

    def patched_tqdm(*args, **kwargs):
        from tqdm import tqdm

        return tqdm(
            *args,
            **kwargs,
            # If true, use ASCII characters ' 123456789#' instead of unicode (smooth blocks) to fill the meter.
            ascii=settings.TQDM_ASCII_LOGGING,
            # Whether to disable the entire progressbar wrapper. If set to None, disable on non-TTY.
            disable=settings.TQDM_DISABLE,
        )

    feature_store.tqdm = patched_tqdm



import dash_ace
import dash_bootstrap_components as dbc
from dash import html
from dash.development.base_component import Component

from amora.models import Model, model_path_for_model


def python_component(model: Model) -> Component:
    source_code = model_path_for_model(model).read_text()
    return html.Div(
        [
            dash_ace.DashAceEditor(
                id="input",
                value=source_code,
                theme="github",
                mode="python",
                tabSize=2,
                enableBasicAutocompletion=True,
                enableLiveAutocompletion=True,
                autocompleter="/autocompleter?prefix=",
                placeholder="Python code ...",
            )
        ]
    )


def sql_component(model: Model) -> Component:
    if model.source() is None:
        return html.Div(
            dbc.Alert(
                "Sourceless model",
                color="primary",
                dismissable=True,
                is_open=True,
            )
        )

    try:
        source_code = model.target_path(
            model_file_path=model_path_for_model(model)
        ).read_text()
    except FileNotFoundError:
        return dbc.Alert(
            "⚠️ SQL code unavailable. Run `amora compile` to generate SQL files",
            color="warning",
        )
    else:
        return html.Div(
            [
                dash_ace.DashAceEditor(
                    id="input",
                    value=source_code,
                    theme="github",
                    mode="SQL",
                    tabSize=2,
                    enableBasicAutocompletion=True,
                    enableLiveAutocompletion=True,
                    autocompleter="/autocompleter?prefix=",
                    placeholder="SQL code ...",
                )
            ]
        )



from typing import Protocol, Union, runtime_checkable

from sqlalchemy.sql import Select, Selectable

Compilable = Union[Select, Selectable]


@runtime_checkable
class CompilableProtocol(Protocol):  # pragma: nocover
    def source(self) -> Compilable:
        ...



from typing import List, Protocol, runtime_checkable

from amora.models import Column


@runtime_checkable
class FeatureViewSourceProtocol(Protocol):  # pragma: nocover
    """
    The contract needed to expose the [AmoraModel](amora-model) definition as a [Feature View](feature-view.md)

    To expose an `AmoraModel` as a Feature View, one must decorate it with
    `amora.feature_store.decorators.feature_view` and implement the
    `FeatureViewProcotol` class methods.

    !!! hint
        An example can be found at `examples.amora_project.models.step_count_by_source.py`

    """

    @classmethod
    def feature_view_entities(cls) -> List[Column]:
        """
        Returns a list of the model columns that should be used as an [entity](feature-view.md#entities). E.g:

        An `AmoraModel` with a single `customer_id` entity:

        ```python
        @classmethod
        def feature_view_entities(cls) -> List[Column]:
            return [cls.customer_id]
        ```

        An `AmoraModel` with multiple entities, `customer_id` and `company_id`:

        ```python
        @classmethod
        def feature_view_entities(cls) -> List[Column]:
            return [cls.customer_id, cls.company_id]
        ```

        A model with no entities:

        ```python
        @classmethod
        def feature_view_entities(cls) -> List[Column]:
            return []
        ```
        """
        ...

    @classmethod
    def feature_view_features(cls) -> List[Column]:
        """
        Returns a list of the model columns that should be used as [features](feature-view.md#features). E.g:

        Features of a customer entity could be _the number of transactions
        they have made on a month_ as `count_transactions_last_30d` and _the sum
        of the transactions amounts they have made on a month_
        as `sum_transactions_last_30d`:

        ```python
        @classmethod
        def feature_view_features(cls) -> List[Column]:
            return [cls.count_transactions_last_30d, cls.sum_transactions_last_30d]
        ```
        """
        ...

    @classmethod
    def feature_view_event_timestamp(cls) -> Column:
        """
        Event timestamp column used for point-in-time joins of feature values. E.g:

        ```python
        @classmethod
        def feature_view_event_timestamp(cls) -> Column:
            return cls.event_timestamp
        ```

        At your Amora Model, the column should be defined as such:

        ```python
        from sqlalchemy import TIMESTAMP, Column

        from amora.models import AmoraModel, Field


        @feature_view
        class FeatureViewModel(AmoraModel, table=True):
            ...
            event_timestamp: datetime = Field(sa_column=Column(TIMESTAMP))
        ```

        """
        ...



import typer

app = typer.Typer(help="Amora dashboards")


@app.command("serve")
def serve():
    from amora.dash.app import dash_app
    from amora.dash.config import settings

    dash_app.run(debug=settings.DEBUG, host=settings.HTTP_HOST, port=settings.HTTP_PORT)



import dash_bootstrap_components as dbc
import plotly.express as px
from dash import dash_table, dcc, html
from dash.development.base_component import Component

from amora.logger import logger
from amora.questions import Question
from amora.visualization import BarChart, BigNumber, LineChart, PieChart


def answer_visualization(question: Question) -> Component:
    try:
        visualization = question.render()
    except Exception:
        logger.exception("Unable to render question answer")
        return dbc.Alert("⚠️ Unable to render question answer", color="warning")
    view_config = visualization.config

    df = visualization.data
    if df.empty:
        return dbc.Alert("⚠️ No data", color="warning")

    if isinstance(view_config, BigNumber):
        big_number = view_config.value_func(df)
        return html.H4(big_number)

    if isinstance(view_config, LineChart):
        return dcc.Graph(
            figure=px.line(
                df,
                x=view_config.x_func(df),
                y=view_config.y_func(df),
            )
        )

    if isinstance(view_config, BarChart):
        return dcc.Graph(
            figure=px.bar(
                df,
                x=view_config.x_func(df),
                y=view_config.y_func(df),
            )
        )

    if isinstance(view_config, PieChart):
        return dcc.Graph(
            figure=px.pie(df, values=view_config.values, names=view_config.names)
        )

    return dash_table.DataTable(
        columns=[
            {"name": col, "id": col, "selectable": True} for col in df.columns.values
        ],
        data=df.to_dict("records"),
        row_selectable="multi",
        sort_action="native",
        style_cell={
            "overflow": "hidden",
            "textOverflow": "ellipsis",
            "maxWidth": 0,
        },
        export_format="csv",
        export_headers="display",
    )


def component(question: Question) -> Component:
    return dbc.Card(
        className="question-card",
        children=dbc.CardBody(
            [
                html.H5(question.name, className="card-title"),
                answer_visualization(question),
                dbc.Accordion(
                    [
                        dbc.AccordionItem(
                            html.Code(question.sql, lang="sql"), title="SQL"
                        )
                    ],
                    start_collapsed=True,
                ),
            ]
        ),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['amora',
 'amora.cli',
 'amora.dash',
 'amora.dash.components',
 'amora.dash.components.filters',
 'amora.dash.pages',
 'amora.feature_store',
 'amora.providers',
 'amora.tests']

package_data = \
{'': ['*'], 'amora': ['target/*', 'templates/*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'SQLAlchemy[mypy]==1.4.39',
 'fsspec>=2022.5.0,<2023.0.0',
 'gcsfs>=2022.5.0,<2023.0.0',
 'humanize>=4.2.3,<5.0.0',
 'matplotlib>=3.4.2,<4.0.0',
 'networkx[all]>=2.6.3,<3.0.0',
 'numpy>=1.21.1,<2.0.0',
 'pandas>=1.3.0,<2.0.0',
 'protobuf<3.20.0',
 'pytest-xdist[psutil]>=2.5.0,<3.0.0',
 'pytest>=6.2.5,<8.0.0',
 'rich>=10.13,<13.0',
 'shed>=0.9.5,<0.11.0',
 'sqlalchemy-bigquery>=1.2,<2.0',
 'sqlmodel>=0.0.4,<0.0.7',
 'sqlparse>=0.3.1,<0.5.0',
 'typer[all]>=0.4,<0.7']

extras_require = \
{'dash': ['Werkzeug==2.1.2',
          'dash[testing]==2.6.1',
          'dash-cytoscape==0.3.0',
          'dash-bootstrap-components>=1.1.0,<2.0.0',
          'Authlib>=1.0',
          'dash-ace>=0.2.1,<0.3.0',
          'dash-extensions>=0.1.5,<0.2.0',
          'dash-mantine-components>=0.10.2,<0.11.0'],
 'feature-store': ['feast[gcp,redis]==0.18.0',
                   'prometheus-fastapi-instrumentator>=5.8.1,<6.0.0']}

entry_points = \
{'console_scripts': ['amora = amora.cli:main'],
 'pytest11': ['amora = amora.tests.pytest_plugin']}

setup_kwargs = {
    'name': 'amora',
    'version': '0.1.11rc18',
    'description': 'Amora Data Build Tool',
    'long_description': 'None',
    'author': 'diogommartins',
    'author_email': 'diogo.martins@stone.com.br',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



import typer

target_option = typer.Option(
    None,
    "--target",
    "-t",
    help="Target connection configuration as defined as an amora.target.Target",
)

models_option = typer.Option(
    [],
    "--model",
    help="A model to be compiled. This option can be passed multiple times.",
)



import sqlparse
from sqlalchemy_bigquery import STRUCT, BigQueryDialect
from sqlalchemy_bigquery.base import BigQueryCompiler

from amora.protocols import Compilable


class AmoraBigQueryCompiler(BigQueryCompiler):
    def visit_getitem_binary(self, binary, operator_, **kwargs):
        left = self.process(binary.left, **kwargs)
        right = self.process(binary.right, **kwargs)

        try:
            # Only integer values should be wrapped in OFFSET
            return f"{left}[OFFSET({int(right)})]"
        except ValueError:
            return f"{left}[{right}]"

    def visit_array(self, element, **kwargs) -> str:
        return "ARRAY[%s]" % self.visit_clauselist(element, **kwargs)

    def visit_struct(self, element, **kwargs) -> str:
        clause_list = self.visit_clauselist(element, **kwargs)
        return f"{element.type.get_col_spec()}{clause_list}"

    def visit_function(self, func, add_to_result_map=None, **kwargs):
        text = super().visit_function(func, add_to_result_map=None, **kwargs)
        if hasattr(func, "_with_offset") and func._with_offset is not None:
            text += f" WITH OFFSET AS {func._with_offset}"
        return text

    def render_literal_value(self, value, type_):
        if isinstance(type_, STRUCT):
            values = ",".join(
                self.render_literal_value(v, type_._STRUCT_byname[k])
                for k, v in value.items()
            )
            return f"({values})"
        return super().render_literal_value(value, type_)


dialect = BigQueryDialect()
dialect.statement_compiler = AmoraBigQueryCompiler


def compile_statement(statement: Compilable) -> str:
    raw_sql = str(
        statement.compile(dialect=dialect, compile_kwargs={"literal_binds": True})
    )
    formatted_sql = sqlparse.format(raw_sql, reindent=True, indent_columns=True)
    return formatted_sql



from enum import Enum
from typing import List, Optional
from uuid import uuid4

import dash_bootstrap_components as dbc
from pydantic import BaseSettings, SecretStr, validator


class DashSettings(BaseSettings):
    HTTP_HOST: str = "127.0.0.1"
    HTTP_PORT: str = "8050"
    DBC_THEME: str = "SKETCHY"
    DEBUG: bool = False

    APP_SECRET_KEY: SecretStr = SecretStr(uuid4().hex)

    AUTH0_CLIENT_ID: Optional[str] = None
    AUTH0_CLIENT_SECRET: Optional[SecretStr] = None
    AUTH0_DOMAIN: Optional[str] = None
    AUTH0_LOGIN_PATH: str = "/login"
    AUTH0_LOGOUT_PATH: str = "/logout"
    AUTH0_CALLBACK_PATH: str = "/callback"
    AUTH0_CALLBACK_REDIRECT_PATH: str = "/"
    AUTH0_SCOPE: str = "openid profile email"

    class Config:
        env_prefix = "AMORA_DASH_"

    @property
    def auth0_login_enabled(self) -> bool:
        return (
            bool(self.AUTH0_CLIENT_ID)
            and bool(self.AUTH0_CLIENT_SECRET)
            and bool(self.AUTH0_DOMAIN)
        )

    @property
    def dbc_theme_stylesheet(self) -> str:
        return getattr(dbc.themes, self.DBC_THEME)

    @property
    def external_stylesheets(self) -> List[str]:
        return [self.dbc_theme_stylesheet, dbc.icons.FONT_AWESOME]

    @validator("DBC_THEME")
    def is_valid_dash_dbc_theme(cls, v):
        assert hasattr(dbc.themes, v)
        return v


settings = DashSettings()


class Size(str, Enum):
    extra_small = "xs"
    small = "sm"
    medium = "md"
    large = "lg"
    extra_large = "xl"


class Color(str, Enum):
    primary = "primary"
    secondary = "secondary"
    success = "success"
    warning = "warning"
    info = "info"
    danger = "danger"



from abc import ABC
from typing import Protocol, Union, runtime_checkable

import pandas as pd
from pandas import Series
from pydantic import BaseModel


@runtime_checkable
class SeriesSelectorFunc(Protocol):
    def __call__(self, df: pd.DataFrame) -> Series:
        pass


@runtime_checkable
class ValueSelectorFunc(Protocol):
    def __call__(self, df: pd.DataFrame) -> str:
        pass


class VisualizationConfig(ABC):
    pass


class PieChart(VisualizationConfig, BaseModel):
    values: str
    names: str


class _2DChart(VisualizationConfig, BaseModel):
    x_func: SeriesSelectorFunc = lambda data: data["x"]
    y_func: SeriesSelectorFunc = lambda data: data["y"]

    class Config:
        arbitrary_types_allowed = True


class BarChart(VisualizationConfig, BaseModel):
    x_func: SeriesSelectorFunc = lambda data: data["x"]
    y_func: SeriesSelectorFunc = lambda data: data["y"]

    class Config:
        arbitrary_types_allowed = True


class LineChart(_2DChart):
    pass


class BigNumber(VisualizationConfig, BaseModel):
    value_func: ValueSelectorFunc = lambda data: data["total"][0]

    class Config:
        arbitrary_types_allowed = True


class Table(VisualizationConfig, BaseModel):
    title: Union[str, None] = None


class Visualization:
    """
    The Amora visual representation of a `pandas.DataFrame`
    """

    def __init__(self, data: pd.DataFrame, config: VisualizationConfig):
        self.data = data
        self.config = config

    def __str__(self):
        return self.data.to_markdown()



from concurrent import futures
from typing import Dict, List, Optional

import pytest
import typer

from amora import materialization
from amora.cli import dash, feature_store, models
from amora.cli.shared_options import models_option, target_option
from amora.cli.type_specs import Models
from amora.compilation import compile_statement
from amora.config import settings
from amora.dag import DependencyDAG
from amora.models import list_models
from amora.utils import list_target_files

app = typer.Typer(
    help="Amora Data Build Tool enables engineers to transform data in their warehouses "
    "by defining schemas and writing select statements with SQLAlchemy. Amora handles turning these "
    "select statements into tables and views"
)


@app.command()
def compile(
    models: Optional[Models] = models_option,
    target: Optional[str] = target_option,
) -> None:
    """
    Generates executable SQL from model files. Compiled SQL files are written to the `./target` directory.
    """
    for model, model_file_path in list_models():
        if models and model_file_path.stem not in models:
            continue

        source_sql_statement = model.source()
        if source_sql_statement is None:
            typer.echo(f"⏭ Skipping compilation of model `{model_file_path}`")
            continue

        target_file_path = model.target_path(model_file_path)
        typer.echo(f"🏗 Compiling model `{model_file_path}` -> `{target_file_path}`")

        content = compile_statement(source_sql_statement)
        target_file_path.parent.mkdir(parents=True, exist_ok=True)
        target_file_path.write_text(content)


@app.command()
def materialize(
    models: Optional[Models] = models_option,
    target: str = target_option,
    draw_dag: bool = typer.Option(False, "--draw-dag"),
    no_compile: bool = typer.Option(
        False,
        "--no-compile",
        help="Don't run `amora compile` before the materialization",
    ),
) -> None:
    """
    Executes the compiled SQL against the current target database.

    """
    if not no_compile:
        compile(models=models, target=target)

    model_to_task: Dict[str, materialization.Task] = {}

    for target_file_path in list_target_files():
        if models and target_file_path.stem not in models:
            continue

        task = materialization.Task.for_target(target_file_path)
        model_to_task[task.model.unique_name()] = task

    dag = DependencyDAG.from_tasks(tasks=model_to_task.values())

    if draw_dag:
        dag.draw()

    with futures.ProcessPoolExecutor(
        max_workers=settings.MATERIALIZE_NUM_THREADS
    ) as executor:
        for models_to_materialize in dag.topological_generations():

            current_tasks: List[materialization.Task] = []
            for model_name in models_to_materialize:
                if model_name in model_to_task:
                    current_tasks.append(model_to_task[model_name])
                else:
                    typer.echo(f"⚠️  Skipping `{model_name}`")
                    continue

            if not current_tasks:
                continue

            results = executor.map(
                materialization.materialize,
                [current_task.sql_stmt for current_task in current_tasks],
                [current_task.model.unique_name() for current_task in current_tasks],
                [current_task.model.__model_config__ for current_task in current_tasks],
            )

            for result in results:
                if result:
                    typer.echo(f"✅  Created `{result.full_table_id}`")
                    typer.echo(f"    Rows: {result.num_rows}")
                    typer.echo(f"    Bytes: {result.num_bytes}")


@app.command()
def test(
    models: Optional[Models] = models_option,
) -> None:
    """
    Runs tests on data in deployed models. Run this after `amora materialize`
    to ensure that the date state is up-to-date.
    """
    return_code = pytest.main(["-n", "auto", "--verbose"])
    raise typer.Exit(return_code)


app.add_typer(dash.app, name="dash")
app.add_typer(models.app, name="models")
app.add_typer(feature_store.app, name="feature-store")



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional


@dataclass
class TimingInfo:
    name: str
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None


@dataclass
class BaseResult:
    """
    Base dataclass for query execution result data
    """

    total_bytes: int
    query: Optional[str]
    job_id: Optional[str]
    referenced_tables: List[str]
    user_email: Optional[str]
